"""Test Recipe against multiple database engines"""

import os
import pytest
from datetime import date, datetime

from sqlalchemy import Column, Date, DateTime, Integer, MetaData, String, Table, insert

from recipe import Recipe, Shelf, get_oven
from recipe.oven.base import OvenBase

from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker


def str_dedent(s):
    return "\n".join([x.lstrip() for x in s.split("\n")]).lstrip("\n")


@pytest.mark.skip("Can't run this without connection")
class TestPostgres(object):
    def setup(self):
        connection_string = os.environ.get("POSTGRES_CONNECTION_STR", None)
        self.skip_tests = False
        if connection_string is None:
            self.skip_tests = True
            return

        self.oven = get_oven(connection_string)

        self.meta = MetaData(bind=self.oven.engine)
        self.session = self.oven.Session()

        self.table = Table(
            "brands", self.meta, autoload=True, autoload_with=self.oven.engine
        )
        d = {"id": {"icon": "check-square", "kind": "Dimension", "field": "id"}}

        self.shelf = self.shelf_from_yaml(
            """
    id:
        kind: Dimension
        field: id
    webflow_brand_profile_id:
        kind: Dimension
        field: webflow_brand_profile_id
    """,
            self.table,
        )

    def shelf_from_yaml(self, yaml_config, selectable):
        """Create a shelf directly from configuration"""
        return Shelf.from_validated_yaml(yaml_config, selectable)

    def recipe(self, **kwargs):
        return Recipe(shelf=self.shelf, session=self.session, **kwargs)

    def testit(self):

        tables = {}
        engine = self.oven.engine
        with engine.connect() as conn:
            for schema in engine.dialect.get_schema_names(conn):
                schema_tables = engine.table_names(schema=schema, connection=conn)
                if schema_tables:
                    tables[schema] = sorted(schema_tables)

        r = self.recipe().dimensions("id")
        assert len(r.all()) == 10


def get_gcloud_credentials_obj():
    # FIXME: Load these creds from a file
    bq_creds = {}
    return bq_creds


def get_bigquery():
    """Get a BigQuery client authenticated with our credentials."""
    google_creds = get_gcloud_credentials_obj()
    return bigquery.Client(credentials=google_creds, project=google_creds.project_id)


class BQOven:
    def __init__(self, connection_string=None):
        self.engine = self.init_engine()
        self.Session = sessionmaker(bind=self.engine)

    def init_engine(self, *args, **kwargs):
        conn_string = "bigquery://juicebox-open-test/juicebox"
        engine_kwargs = {"credentials_info": get_gcloud_credentials_obj()}
        return create_engine(conn_string, **engine_kwargs)


@pytest.mark.skip("Can't run this witout connection")
class TestBigQuery(object):
    def setup(self):
        self.oven = BQOven()
        self.meta = MetaData(bind=self.oven.engine)
        self.session = self.oven.Session()

        self.table = Table(
            "runningwithtimestamp_8599_b3310438d33748de",
            self.meta,
            autoload=True,
            autoload_with=self.oven.engine,
        )
        d = {"id": {"icon": "check-square", "kind": "Dimension", "field": "id"}}

        self.shelf = self.shelf_from_yaml(
            """
    date:
        kind: Dimension
        field: day(date)
    name:
        kind: Dimension
        field: name
    distance:
        kind: Metric
        field: sum(distance)

    """,
            self.table,
        )

    def shelf_from_yaml(self, yaml_config, selectable):
        """Create a shelf directly from configuration"""
        return Shelf.from_validated_yaml(yaml_config, selectable)

    def recipe(self, **kwargs):
        return Recipe(shelf=self.shelf, session=self.session, **kwargs)

    def test_query(self):
        from datetime import datetime

        d1 = datetime(2020, 1, 1)
        d2 = datetime(2020, 12, 31)
        f = self.shelf["date"].build_filter(
            operator="between", value=["2020-01-01", "2020-12-31"]
        )
        f2 = self.shelf["date"].build_filter(operator="between", value=[d1, d2])
        for c in self.table.columns:
            print(c, c.type)
        print(f)
        print(f2)
        recipe = self.recipe().dimensions("date").metrics("distance").filters(f2)
        print(recipe.to_sql())

        for row in recipe.all():
            print(row)
        # assert 1 == 2

    def testit(self):
        print(self.table)
        for c in self.table.columns:
            print(c, c.type)
        assert 1 == 1


@pytest.mark.skip("Can't run this witout connection")
class TestRecipeSQLServer(object):
    def setup(self):
        connection_string = os.environ.get("SQL_SERVER_CONNECTION_STR", None)
        self.skip_tests = False
        if connection_string is None:
            self.skip_tests = True
            return

        self.oven = get_oven(connection_string)

        self.meta = MetaData(bind=self.oven.engine)
        self.session = self.oven.Session()

        self.table = Table(
            "foo",
            self.meta,
            Column("first", String),
            Column("last", String),
            Column("age", Integer),
            Column("birth_date", Date),
            Column("dt", DateTime),
            extend_existing=True,
        )
        self.meta.create_all(self.oven.engine)

        data = [
            {
                "first": "hi",
                "last": "there",
                "age": 5,
                "birth_date": date(2015, 1, 1),
                "dt": datetime(2005, 12, 1, 12, 15),
            },
            {
                "first": "hi",
                "last": "fred",
                "age": 10,
                "birth_date": date(2015, 5, 15),
                "dt": datetime(2013, 10, 15, 5, 20, 10),
            },
        ]
        with self.oven.engine.connect() as conn:
            for row in data:
                conn.execute(insert(self.table).values(**row))

        self.shelf = self.shelf_from_yaml(
            """
first:
    kind: Dimension
    field: first
last:
    kind: Dimension
    field: last
firstlast:
    kind: Dimension
    field: "first + last"
    id_field: first
age:
    kind: Measure
    field: sum(age)
test_month:
    kind: Dimension
    field: month(birth_date)
year_by_format:
    kind: Dimension
    field: dt
    format: "%Y"
count:
    kind: Measure
    field: count(*)
""",
            self.table,
        )

    def teardown(self):
        if not self.skip_tests:
            self.meta.drop_all(self.oven.engine)

    def shelf_from_yaml(self, yaml_config, selectable):
        """Create a shelf directly from configuration"""
        return Shelf.from_validated_yaml(yaml_config, selectable)

    def recipe(self, **kwargs):
        return Recipe(shelf=self.shelf, session=self.session, **kwargs)

    def assertRecipeCSV(self, recipe, content):
        actual = recipe.dataset.csv.replace("\r\n", "\n")
        expected = str_dedent(content)
        assert actual == expected

    def test_dimension(self):
        if self.skip_tests:
            return
        recipe = self.recipe().metrics("age", "count").dimensions("first")
        self.assertRecipeCSV(
            recipe,
            """
            first,age,count,first_id
            hi,15,2,hi
            """,
        )
        recipe = self.recipe().metrics("age").dimensions("firstlast")
        self.assertRecipeCSV(
            recipe,
            """
            firstlast_id,firstlast,age,firstlast_id
            hi,hifred,10,hi
            hi,hithere,5,hi
            """,
        )

    def test_dates_and_Datetimes(self):
        """We can convert dates using formats"""
        if self.skip_tests:
            return
        recipe = (
            self.recipe()
            .dimensions("year_by_format")
            .metrics("count")
            .order_by("year_by_format")
        )
        self.assertRecipeCSV(
            recipe,
            """
            year_by_format,count,year_by_format_id
            2005-01-01 00:00:00,1,2005-01-01 00:00:00
            2013-01-01 00:00:00,1,2013-01-01 00:00:00
            """,
        )
        recipe = (
            self.recipe()
            .dimensions("year_by_format")
            .metrics("count")
            .order_by("-year_by_format")
        )
        self.assertRecipeCSV(
            recipe,
            """
            year_by_format,count,year_by_format_id
            2013-01-01 00:00:00,1,2013-01-01 00:00:00
            2005-01-01 00:00:00,1,2005-01-01 00:00:00
            """,
        )

        # Test a month() conversion
        recipe = (
            self.recipe()
            .dimensions("test_month")
            .metrics("age", "count")
            .order_by("-test_month")
        )
        self.assertRecipeCSV(
            recipe,
            """
            test_month,age,count,test_month_id
            2015-05-01,10,1,2015-05-01
            2015-01-01,5,1,2015-01-01
            """,
        )



# -*- coding: UTF-8 -*-
from recipe.utils import clean_unicode
from unittest import TestCase


class TestCleanUnicode(TestCase):
    def test_string(self):
        test_values = [
            ("cookies", b"cookies"),
            (22.04, b"22.04"),
            (None, b"None"),
            (
                "Falsches Üben von Xylophonmusik quält jeden größeren Zwerg",
                b"Falsches Uben von Xylophonmusik qualt jeden groeren Zwerg",
            ),
            ("«küßî»", b"kui"),
            # String dashes
            ("― Like this? ― Right.", b" Like this?  Right."),
            ("oòóôõöōŏǫȯőǒȍȏ", b"oooooooooooooo"),
            # two byte no chars
            ("“ЌύБЇ”", b""),
        ]
        for actual, expected in test_values:
            value = clean_unicode(actual)
            self.assertEqual(value, expected)



from uuid import uuid4

import sqlparse
from sqlalchemy.engine.default import DefaultDialect
from sqlalchemy.exc import UnsupportedCompilationError
from sqlalchemy.orm import Query
from sqlalchemy.sql.sqltypes import Date, DateTime, NullType, String


def expr_to_str(expr):
    """Utility to print sql for a expression"""
    return str(expr.compile(compile_kwargs={"literal_binds": True}))


def filter_to_string(filt):
    """Compile a filter object to a literal string"""
    try:
        if hasattr(filt, "filters") and filt.filters:
            return expr_to_str(filt.filters[0])
        elif hasattr(filt, "havings") and filt.havings:
            return expr_to_str(filt.havings[0])
        elif isinstance(filt, bool):
            return str(filt)
        else:
            return expr_to_str(filt)
    except UnsupportedCompilationError:
        return uuid4()


class StringLiteral(String):
    """Teach SA how to literalize various things."""

    def literal_processor(self, dialect):
        super_processor = super(StringLiteral, self).literal_processor(dialect)

        def process(value):
            if isinstance(value, int):
                return str(value)
            if not isinstance(value, str):
                value = str(value)
            result = super_processor(value)
            if isinstance(result, bytes):
                result = result.decode(dialect.encoding)
            return result

        return process


def prettyprintable_sql(statement, dialect=None, reindent=True):
    """
    Generate an SQL expression string with bound parameters rendered inline
    for the given SQLAlchemy statement. The function can also receive a
    `sqlalchemy.orm.Query` object instead of statement.

    WARNING: Should only be used for debugging. Inlining parameters is not
             safe when handling user created data.
    """
    if isinstance(statement, Query):
        if dialect is None:
            dialect = statement.session.get_bind().dialect
        statement = statement.statement

    # Generate a class that can handle encoding
    if dialect:
        DialectKlass = dialect.__class__
    else:
        DialectKlass = DefaultDialect

    class LiteralDialect(DialectKlass):
        colspecs = {
            # prevent various encoding explosions
            String: StringLiteral,
            # teach SA about how to literalize a datetime
            DateTime: StringLiteral,
            Date: StringLiteral,
            # don't format py2 long integers to NULL
            NullType: StringLiteral,
        }

    compiled = statement.compile(
        dialect=LiteralDialect(), compile_kwargs={"literal_binds": True}
    )
    return sqlparse.format(str(compiled), reindent=reindent)



import contextlib
from collections import OrderedDict
from copy import copy
from dataclasses import dataclass, field
from typing import Optional, List, Dict

from lark.exceptions import VisitError
from sqlalchemy import Float, Integer, String, Table
from sqlalchemy.util import lightweight_named_tuple
from sureberus import errors as E
from sureberus import normalize_schema
from yaml import safe_load

from recipe.exceptions import BadIngredient, BadRecipe, InvalidColumnError
from recipe.ingredients import Dimension, Filter, Ingredient, InvalidIngredient, Metric
from recipe.schemas import shelf_schema
from recipe.schemas.builders import SQLAlchemyBuilder
from recipe.schemas.parsed_constructors import create_ingredient_from_parsed

_POP_DEFAULT = object()


# FIXME: we can do all this in create_ingredient_from_parsed; we don't need this layer
def ingredient_from_validated_dict(ingr_dict, selectable, builder=None):
    """Create an ingredient object from a validated ingredient schema"""
    try:
        return create_ingredient_from_parsed(ingr_dict, builder)
    except VisitError as e:
        # Lark returns the InvalidColumnError wrapped in a VisitError
        if isinstance(e.orig_exc, InvalidColumnError):
            # custom exception handling
            error = {
                "type": "invalid_column",
                "extra": {"column_name": e.orig_exc.column_name},
            }
            return InvalidIngredient(error=error)
        else:
            raise


@dataclass
class SelectParts:
    columns: list = field(default_factory=list)
    group_bys: list = field(default_factory=list)
    filters: set = field(default_factory=set)
    havings: set = field(default_factory=set)
    raw_order_by_keys: list = field(default_factory=list)
    order_bys: list = field(default_factory=list)
    all_filters: set = field(default_factory=set)

    def add_ingredient(self, ingredient):
        """Gather the SQLAlchemy fragments from this ingredient into a consolidated list."""
        if ingredient.error:
            error_type = ingredient.error.get("type")
            if error_type == "invalid_column":
                extra = ingredient.error.get("extra", {})
                column_name = extra.get("column_name")
                ingredient_name = extra.get("ingredient_name")
                error_msg = 'Invalid column "{0}" in ingredient "{1}"'.format(
                    column_name, ingredient_name
                )
                raise InvalidColumnError(error_msg, column_name=column_name)
            raise BadIngredient(str(ingredient.error))
        self.columns.extend(ingredient.labeled_columns)
        self.group_bys.extend(ingredient.group_by)
        self.havings.update(ingredient.havings)
        if ingredient.filters:
            # Ensure we don't add duplicate filters
            for new_f in ingredient.filters:
                from recipe.utils import filter_to_string

                new_f_str = filter_to_string(new_f)
                if new_f_str not in self.all_filters:
                    self.filters.add(new_f)
                    self.all_filters.add(new_f_str)

        # Hoist any order by into the ordering
        if (
            "order_by" in ingredient.roles
            and ingredient.id not in self.raw_order_by_keys
            and f"-{ingredient.id}" not in self.raw_order_by_keys
        ):
            if ingredient.ordering == "desc":
                self.raw_order_by_keys.append(f"-{ingredient.id}")
            else:
                self.raw_order_by_keys.append(ingredient.id)

    def validate_order_bys(self, shelf):
        validated_order_bys = OrderedDict()
        for key in self.raw_order_by_keys:
            with contextlib.suppress(BadRecipe):
                ingr = shelf.find(key, (Dimension, Metric))
                for c in ingr.order_by_columns(engine=shelf.Meta.engine):
                    # Avoid duplicate order by columns
                    if str(c) not in [str(o) for o in validated_order_bys]:
                        validated_order_bys[c] = None
        self.order_bys = list(validated_order_bys.keys())


class Shelf(object):
    """Holds ingredients used by a recipe.

    Can be initialized with no arguments, but also accepts:
    - a dictionary of ingredients as a positional argument
    - ingredients as keyword arguments

    These keyword arguments have special meaning:

    :param select_from: The SQLALchemy-compatible object which
        will be queried (usually a Table or ORM object).
    :param table: Unused, but stored on the `Meta` attribute.
    :param metadata: Unused, but stored on the `Meta` attribute.
    """

    class Meta:
        anonymize = False
        table = None
        select_from = None
        ingredient_order = []
        metadata = None
        engine = None

    def __init__(self, *args, **kwargs):
        self.Meta = type(self).Meta()
        self.Meta.ingredient_order = []
        self.Meta.table = kwargs.pop("table", None)
        self.Meta.select_from = kwargs.pop("select_from", None)
        self.Meta.metadata = kwargs.pop("metadata", None)
        self.Meta.engine = kwargs.pop("engine", None)
        self._ingredients = {}
        self.update(*args, **kwargs)

    # Dict Interface

    def get(self, k, d=None):
        ingredient = self._ingredients.get(k, d)
        if isinstance(ingredient, Ingredient):
            ingredient.id = k
            ingredient.anonymize = self.Meta.anonymize
        return ingredient

    def items(self):
        """Return an iterator over the ingredient names and values."""
        return self._ingredients.items()

    def values(self):
        """Return an iterator over the ingredients."""
        return self._ingredients.values()

    def keys(self):
        """Return an iterator over the ingredient keys."""
        return self._ingredients.keys()

    def __copy__(self):
        meta = copy(self.Meta)
        ingredients = copy(self._ingredients)
        new_shelf = type(self)(ingredients)
        new_shelf.Meta = meta
        return new_shelf

    def __iter__(self):
        return iter(self._ingredients)

    def __getitem__(self, key):
        """Set the id and anonymize property of the ingredient whenever we
        get or set items"""
        ingr = self._ingredients[key]
        # Ensure the ingredient's `anonymize` matches the shelf.

        # TODO: this is nasty, but *somewhat* safe because we are (hopefully)
        # guaranteed to "own" copies of all of our ingredients. It would be
        # much better if Shelf had logic that ran when anonymize is set to
        # update all ingredients. Or better yet, the code that anonymizes
        # queries should just look at the shelf instead of the ingredients.

        # One way in this is "spooky" is:
        # ingr = shelf['foo']
        # # ingr.anonymize is now False
        # shelf.Meta.anonymize = True
        # # ingr.anonymize is still False
        # shelf['foo] # ignore result
        # # ingr.anonymize is now True

        ingr.anonymize = self.Meta.anonymize
        return ingr

    def __setitem__(self, key, ingredient):
        """Set the id and anonymize property of the ingredient whenever we
        get or set items"""
        # Maintainer's note: try to make all mutation of self._ingredients go
        # through this method, so we can reliably copy & annotate the
        # ingredients that go into the Shelf.
        if not isinstance(ingredient, Ingredient):
            raise TypeError(
                "Can only set Ingredients as items on Shelf. "
                "Got: {!r}".format(ingredient)
            )
        ingredient_copy = copy(ingredient)
        ingredient_copy.id = key
        ingredient_copy.anonymize = self.Meta.anonymize
        self._ingredients[key] = ingredient_copy

    def __contains__(self, key):
        return key in self._ingredients

    def __len__(self):
        return len(self._ingredients)

    def clear(self):
        self._ingredients.clear()

    def update(self, d=None, **kwargs):
        items = []
        if d is not None:
            items = list(d.items())
        for k, v in items + list(kwargs.items()):
            self[k] = v

    def pop(self, k, d=_POP_DEFAULT):
        """Pop an ingredient off of this shelf."""
        if d is _POP_DEFAULT:
            return self._ingredients.pop(k)
        else:
            return self._ingredients.pop(k, d)

    # End dict interface

    def ingredients(self):
        """Return the ingredients in this shelf in a deterministic order"""
        return sorted(list(self.values()))

    @property
    def dimension_ids(self):
        """Return the Dimensions on this shelf in the order in which
        they were used."""
        return self._sorted_ingredients(
            [d.id for d in self.values() if isinstance(d, Dimension)]
        )

    @property
    def metric_ids(self):
        """Return the Metrics on this shelf in the order in which
        they were used."""
        return self._sorted_ingredients(
            [d.id for d in self.values() if isinstance(d, Metric)]
        )

    @property
    def filter_ids(self):
        """Return the Filters on this shelf in the order in which
        they were used."""
        return self._sorted_ingredients(
            [d.id for d in self.values() if isinstance(d, Filter)]
        )

    def _sorted_ingredients(self, ingredients):
        def sort_key(id):
            if id in self.Meta.ingredient_order:
                return self.Meta.ingredient_order.index(id)
            else:
                return 9999

        return tuple(sorted(ingredients, key=sort_key))

    def __repr__(self):
        """A string representation of the ingredients used in a recipe
        ordered by Dimensions, Metrics, Filters, then Havings
        """
        lines = [ingredient.describe() for ingredient in sorted(self.values())]
        return "\n".join(lines)

    def use(self, ingredient):
        if not isinstance(ingredient, Ingredient):
            raise TypeError(
                "Can only set Ingredients as items on Shelf. "
                "Got: {!r}".format(ingredient)
            )

        # Track the order in which ingredients are added.
        self.Meta.ingredient_order.append(ingredient.id)
        self[ingredient.id] = ingredient

    @classmethod
    def from_config(
        cls,
        obj: Dict,
        selectable,
        ingredient_constructor=None,
        metadata=None,
        *,
        builder: Optional[SQLAlchemyBuilder] = None,
        ingredient_cache=None,
        extra_selectables: Optional[List] = None,
        constants: Optional[Dict] = None,
    ):
        """Create a shelf using a dict shelf definition.

        :param obj: A Python dictionary describing the configuration of a Shelf.
        :param selectable: A SQLAlchemy Table, a Recipe, a table name, or a
            SQLAlchemy join to select from.
        :param ingredient_constructor: DEPRECATED, a callable used to create
            ingredients
        :param metadata: If `selectable` is passed as a table name, then in
            order to introspect its schema, we must have the SQLAlchemy
            MetaData object to associate it with.
        :param ingredient_cache: An optional cache for improving parse times
        :param extra_selectables: A list of (selectable, namespace) tuples.
            these are extra selectables that can be used in expressions
        :return: A shelf that contains the ingredients defined in obj.
        """

        try:
            validated_shelf = normalize_schema(shelf_schema, obj, allow_unknown=True)
        except E.SureError as e:
            raise BadIngredient(str(e))

        d = {}
        if builder is None:
            from recipe import Recipe

            constants = constants or {}

            if isinstance(selectable, Recipe):
                selectable = selectable.subquery()
            elif isinstance(selectable, str):
                if "." in selectable:
                    schema, tablename = selectable.split(".")
                else:
                    schema, tablename = None, selectable

                selectable = Table(
                    tablename,
                    metadata,
                    schema=schema,
                    extend_existing=True,
                    autoload=True,
                )

            builder = SQLAlchemyBuilder.get_builder(
                selectable=selectable,
                cache=ingredient_cache,
                extra_selectables=extra_selectables,
                constants=constants,
            )

        for k, v in validated_shelf.items():
            d[k] = ingredient_from_validated_dict(v, selectable, builder=builder)

            if isinstance(d[k], InvalidIngredient):
                if not d[k].error.get("extra"):
                    d[k].error["extra"] = {}
                d[k].error["extra"]["ingredient_name"] = k

        engine = builder.selectable.bind

        # TODO: Evaluate how and if we're using select_from
        shelf = cls(d, select_from=builder.selectable, engine=engine)
        if builder and ingredient_cache is not None:
            builder.save_cache()

        return shelf

    @classmethod
    def from_yaml(cls, yaml_str, selectable, **kwargs):
        """Shim that calls from_validated_yaml.

        This used to call a different implementation of yaml parsing
        """
        return cls.from_validated_yaml(yaml_str, selectable, **kwargs)

    @classmethod
    def from_validated_yaml(cls, yaml_str, selectable, **kwargs):
        """Create a shelf using a yaml shelf definition.

        :param yaml_str: A string containing yaml ingredient definitions.
        :param selectable: A SQLAlchemy Table, a Recipe, or a SQLAlchemy
            join to select from.
        :return: A shelf that contains the ingredients defined in yaml_str.
        """
        obj = safe_load(yaml_str)
        return cls.from_config(obj, selectable, **kwargs)

    def find(self, obj, filter_to_class=Ingredient, constructor=None):
        """
        Find an Ingredient, optionally using the shelf.

        :param obj: A string or Ingredient
        :param filter_to_class: The Ingredient subclass that obj must be an
         instance of
        :param constructor: An optional callable for building Ingredients
         from obj
        :return: An Ingredient of subclass `filter_to_class`
        """
        if callable(constructor):
            obj = constructor(obj, shelf=self)

        if isinstance(obj, str):
            set_descending = obj.startswith("-")
            if set_descending:
                obj = obj[1:]

            if obj not in self:
                raise BadRecipe("{} doesn't exist on the shelf".format(obj))

            ingredient = self[obj]
            if isinstance(ingredient, InvalidIngredient):
                # allow InvalidIngredient, it will be handled at a later time
                return ingredient

            if not isinstance(ingredient, filter_to_class):
                raise BadRecipe("{} is not a {}".format(obj, filter_to_class))

            if set_descending:
                ingredient.ordering = "desc"

            return ingredient
        elif isinstance(obj, filter_to_class):
            return obj
        else:
            raise BadRecipe("{} is not a {}".format(obj, filter_to_class))

    def brew_select_parts(self, order_by_keys=None) -> SelectParts:
        if order_by_keys is None:
            order_by_keys = []
        parts = SelectParts(raw_order_by_keys=order_by_keys)

        for ingredient in self.ingredients():
            parts.add_ingredient(ingredient)

        parts.validate_order_bys(self)
        return parts

    def brew_query_parts(self, order_by_keys=None):
        """Make columns, group_bys, filters, havings"""
        columns, group_bys, filters, havings = [], [], set(), set()
        if order_by_keys is None:
            order_by_keys = []
        order_by_keys = list(order_by_keys)
        all_filters = set()

        for ingredient in self.ingredients():
            if ingredient.error:
                error_type = ingredient.error.get("type")
                if error_type == "invalid_column":
                    extra = ingredient.error.get("extra", {})
                    column_name = extra.get("column_name")
                    ingredient_name = extra.get("ingredient_name")
                    error_msg = 'Invalid column "{0}" in ingredient "{1}"'.format(
                        column_name, ingredient_name
                    )
                    raise InvalidColumnError(error_msg, column_name=column_name)
                raise BadIngredient(str(ingredient.error))
            columns.extend(ingredient.labeled_columns)
            group_bys.extend(ingredient.group_by)
            # Ensure we don't add duplicate filters
            for new_f in ingredient.filters:
                from recipe.utils import filter_to_string

                new_f_str = filter_to_string(new_f)
                if new_f_str not in all_filters:
                    filters.add(new_f)
                    all_filters.add(new_f_str)
            havings.update(ingredient.havings)

            # If there is an order_by key on one of the ingredients, make sure
            # the recipe orders by this ingredient
            if "order_by" in ingredient.roles:
                if (
                    ingredient.id not in order_by_keys
                    and "-" + ingredient.id not in order_by_keys
                ):
                    if ingredient.ordering == "desc":
                        order_by_keys.append("-" + ingredient.id)
                    else:
                        order_by_keys.append(ingredient.id)

        order_bys = OrderedDict()
        for key in order_by_keys:
            try:
                ingr = self.find(key, (Dimension, Metric))
                for c in ingr.order_by_columns(engine=self.Meta.engine):
                    # Avoid duplicate order by columns
                    if str(c) not in [str(o) for o in order_bys]:
                        order_bys[c] = None
            except BadRecipe as e:
                # Ignore order_by if the dimension/metric is not used.
                # TODO: Add structlog warning
                pass

        return {
            "columns": columns,
            "group_bys": group_bys,
            "filters": filters,
            "havings": havings,
            "order_bys": list(order_bys.keys()),
        }

    def enchant(self, data, cache_context=None):
        """Add any calculated values to each row of a resultset generating a
        new namedtuple

        :param data: a list of row results
        :param cache_context: optional extra context for caching
        :return: a list with ingredient.cauldron_extras added for all
                 ingredients
        """
        enchantedlist = []
        if data:
            sample_item = data[0]

            # Extra fields to add to each row
            # With extra callables
            extra_fields, extra_callables = [], []

            for ingredient in self.ingredients():
                if not isinstance(ingredient, (Dimension, Metric)):
                    continue
                if cache_context:
                    ingredient.cache_context += str(cache_context)
                for extra_field, extra_callable in ingredient.cauldron_extras:
                    extra_fields.append(extra_field)
                    extra_callables.append(extra_callable)

            # Mixin the extra fields
            keyed_tuple = lightweight_named_tuple(
                "result", sample_item._fields + tuple(extra_fields)
            )

            # Iterate over the results and build a new namedtuple for each row
            for row in data:
                values = row + tuple(fn(row) for fn in extra_callables)
                enchantedlist.append(keyed_tuple(values))

        return enchantedlist


def AutomaticShelf(table):
    """Given a SQLAlchemy Table, automatically generate a Shelf with metrics
    and dimensions based on its schema.
    """
    if hasattr(table, "__table__"):
        table = table.__table__
    config = introspect_table(table)
    return Shelf.from_config(config, table)


def introspect_table(table):
    """Given a SQLAlchemy Table object, return a Shelf description suitable
    for passing to Shelf.from_config.
    """
    d = {}
    for c in table.columns:
        if isinstance(c.type, String):
            d[c.name] = {"kind": "Dimension", "field": c.name}
        if isinstance(c.type, (Integer, Float)):
            d[c.name] = {"kind": "Metric", "field": c.name}
    return d



import hashlib
import inspect
from copy import copy
from datetime import date, datetime

import dateparser
from dateutil.relativedelta import relativedelta
from sqlalchemy import text
from sqlalchemy.ext.declarative import DeclarativeMeta
from sqlalchemy.sql.base import ColumnCollection

ColumnCollection
from sureberus import schema as S

from recipe.exceptions import InvalidColumnError

SCALAR_TYPES = [S.Integer(), S.String(), S.Float(), S.Boolean()]


# SQL server can not support parameters in queries that are used for grouping
# https://github.com/mkleehammer/pyodbc/issues/479
# To avoid parameterization, we pass literals
literal_1 = text("1")
literal_0 = text("0")


def _chain(*args):
    """Chain several coercers together"""

    def fn(value):
        for arg in args:
            value = arg(value)
        return value

    return fn


def _make_sqlalchemy_datatype_lookup():
    """Build a dictionary of the allowed sqlalchemy casts"""
    from sqlalchemy.sql import sqltypes

    d = {}
    for name in dir(sqltypes):
        sqltype = getattr(sqltypes, name)
        if name.lower() not in d and name[0] != "_" and name != "NULLTYPE":
            if inspect.isclass(sqltype) and issubclass(sqltype, sqltypes.TypeEngine):
                d[name.lower()] = sqltype
    return d


sqlalchemy_datatypes = _make_sqlalchemy_datatype_lookup()

format_lookup = {
    "comma": ",.0f",
    "dollar": "$,.0f",
    "percent": ".0%",
    "comma1": ",.1f",
    "comma2": ",.2f",
    "comma3": ",.3f",
    "dollar1": "$,.1f",
    "dollar2": "$,.2f",
    "percent1": ".1%",
    "percent2": ".2%",
    "percent3": ".3%",
}


def coerce_format(v):
    return format_lookup.get(v, v)


def generate_lookup_by_engine(lookup_by_engine, engine):
    """Convert `aggregations_by_engine` or `conversions_by_engine`
    into a dictionary specific to the provided engine/drivername.
    """
    result = copy(lookup_by_engine.get("default", {}))
    result.update(lookup_by_engine.get(engine, {}))
    for k, v in list(result.items()):
        if v is None:
            result.pop(k, None)
    return result


def convert_by_engine_keys_to_regex(lookup_by_engine):
    """Convert all the keys in a lookup_by_engine to a regex"""
    keys = set()
    for d in lookup_by_engine.values():
        for k in d.keys():
            if isinstance(k, str):
                keys.add(k)

    keys = list(keys)
    keys.sort(key=lambda item: (len(item), item), reverse=True)
    return "|".join(keys)


def coerce_pop_version(shelf):
    shelf.pop("_version", None)
    return shelf


def coerce_shelf_meta(shelf):
    """Move shelf meta from the shelf to all ingredients"""
    shelf_meta = shelf.pop("_meta", None)
    for k, v in shelf.items():
        # Add shelf meta to all ingredient definitions
        if shelf_meta and not k.startswith("_"):
            v["_shelf_meta"] = copy(shelf_meta)
    return shelf


def _convert_date_value(v):
    parse_kwargs = {"languages": ["en"]}
    if isinstance(v, date):
        return v
    elif isinstance(v, datetime):
        return v.date()
    elif isinstance(v, str):
        parsed_dt = dateparser.parse(v, **parse_kwargs)
        if parsed_dt is None:
            raise ValueError("Could not parse date in {}".format(v))
        return parsed_dt.date()
    else:
        raise ValueError("Can not convert {} to date".format(v))


def _convert_datetime_value(v):
    parse_kwargs = {"languages": ["en"]}
    if isinstance(v, datetime):
        return v
    elif isinstance(v, date):
        return datetime(v.year, v.month, v.day)
    elif isinstance(v, str):
        parsed_dt = dateparser.parse(v, **parse_kwargs)
        if parsed_dt is None:
            raise ValueError("Could not parse datetime in {}".format(v))
        return parsed_dt
    else:
        raise ValueError("Can not convert {} to datetime".format(v))


def convert_value(field, value):
    """Convert values into something appropriate for this SQLAlchemy data type

    :param field: A SQLAlchemy expression
    :param values: A value or list of values
    """

    if isinstance(value, (list, tuple)):
        if str(field.type) == "DATE":
            return [_convert_date_value(v) for v in value]
        elif str(field.type) == "DATETIME":
            return [_convert_datetime_value(v) for v in value]
        else:
            return value
    else:
        if str(field.type) == "DATE":
            return _convert_date_value(value)
        elif str(field.type) == "DATETIME":
            return _convert_datetime_value(value)
        else:
            return value


def _find_in_columncollection(columns, name):
    """Find a column in a column collection by name or _label"""
    for col in columns:
        if col.name == name or getattr(col, "_label", None) == name:
            return col
    return None


def find_column(selectable, name):
    """
    Find a column named `name` in selectable

    :param selectable:
    :param name:
    :return: A column object
    """
    from recipe import Recipe

    if isinstance(selectable, Recipe):
        selectable = selectable.subquery()

    # Selectable is a table
    if isinstance(selectable, DeclarativeMeta):
        col = getattr(selectable, name, None)
        if col is not None:
            return col

        col = _find_in_columncollection(selectable.__table__.columns, name)
        if col is not None:
            return col

    # Selectable is a sqlalchemy subquery
    elif hasattr(selectable, "c") and isinstance(selectable.c, ColumnCollection):
        col = getattr(selectable.c, name, None)
        if col is not None:
            return col

        col = _find_in_columncollection(selectable.c, name)
        if col is not None:
            return col

    raise InvalidColumnError(column_name=name)


def ingredient_class_for_name(class_name):
    """Get the class in the recipe.ingredients module with the given name."""
    from recipe import ingredients

    return getattr(ingredients, class_name, None)


def date_offset(dt, offset, **offset_params):
    if offset in ("prior", "previous", "last"):
        dt -= relativedelta(**offset_params)
        return dt
    elif offset == "next":
        dt += relativedelta(**offset_params)
        return dt
    elif offset in ("current", "this"):
        return dt
    else:
        raise ValueError("Unknown intelligent date offset")


def convert_to_start_datetime(dt):
    """Convert a date or datetime to the first moment of the day"""
    # Convert a date or datetime to the first moment of the day
    # only if the datetime is the first moment of the day
    if isinstance(dt, (date, datetime)):
        dt = datetime(dt.year, dt.month, dt.day)
    return dt


def convert_to_end_datetime(dt):
    """Convert a date or datetime to the last moment of the day"""
    if isinstance(dt, (date, datetime)):
        dt = datetime(dt.year, dt.month, dt.day)
        dt += relativedelta(days=1, microseconds=-1)
    return dt


def convert_to_eod_datetime(dt):
    """Convert a date or datetime to the last moment of the day,
    only convert datetimes if they are the first moment of the day,"""
    if isinstance(dt, datetime):
        if dt.hour == 0 and dt.minute == 0 and dt.second == 0:
            dt += relativedelta(days=1, microseconds=-1)
    elif isinstance(dt, date):
        dt = datetime(dt.year, dt.month, dt.day)
        dt += relativedelta(days=1, microseconds=-1)
    return dt


def calc_date_range(offset, units, dt):
    """Create an intelligent date range using offsets, units and a starting date

    Args:

        offset: An offset
            current|this options for the current period
            prior|previous|last options for the previous period
            next options the next period
        units: The kind of date range to create
            year: A full year
            ytd: The year up to the provided date
            qtr: The full quarter the date belongs to
            month: The full month of the provided date
            mtd: The month up to the provided date
            day: The provided date
        dt
            The date that will be used for calculations

    Returns:

        A tuple of dates constructed using the offsets and units
    """
    offset = str(offset).lower()
    units = str(units).lower()

    # TODO: Add a week unit
    if units == "year":
        dt = date_offset(dt, offset, years=1)
        return date(dt.year, 1, 1), date(dt.year, 12, 31)
    elif units == "ytd":
        dt = date_offset(dt, offset, years=1)
        return date(dt.year, 1, 1), dt
    elif units == "qtr":
        dt = date_offset(dt, offset, months=3)
        qtr = (dt.month - 1) // 3  # Calculate quarter as 0,1,2,3
        return (
            date(dt.year, qtr * 3 + 1, 1),
            date(dt.year, qtr * 3 + 3, 1) + relativedelta(months=1, days=-1),
        )
    elif units == "month":
        dt = date_offset(dt, offset, months=1)
        start_dt = date(dt.year, dt.month, 1)
        return start_dt, start_dt + relativedelta(months=1, days=-1)
    elif units == "mtd":
        dt = date_offset(dt, offset, months=1)
        return date(dt.year, dt.month, 1), dt
    elif units == "day":
        dt = date_offset(dt, offset, days=1)
        return dt, dt
    else:
        raise ValueError("Unknown intelligent date units")


def mkkey(prefix, *parts):
    hash = hashlib.sha1()
    for k in parts:
        hash.update(str(k).encode("utf-8"))
    return f"{prefix}-{hash.hexdigest()}"
